/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-02-23 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"
#include "soil/SOIL.h"

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"
#include "soil/SOIL.c"
#include "soil/stb_image_aug.c"

/* One simulated particle */
class particle {
public:
	vec4 color; // RGBA
	vec3 pos, vel; // in world coordinates-- meters
	vec3 netforce; // in newtons (quarter-pounds)
	float mass; // in kilograms
	
	void start_step(double dt) {
		/*
		 For stability, I'm using "Leapfrog" integration:
			http://en.wikipedia.org/wiki/Leapfrog_integration
		 Basically, we just update position, *then* calculate
		 net force before moving the velocity.
		*/
		pos+=dt*vel;
		
		/* Zero out net force, so springs can add to it */
		netforce=vec3(0.0);
	}
	
	void finish_step(double dt) {
	/* Newtonian motion */
		vec3 acc=netforce/mass; 

		acc.z+=-9.8;  /* gravity */
		
		/* "wind drag": fairly stiff */
		acc+=-1.2*vel;
		vel+=dt*acc;

		if (pos.z<0.0) { /* hit the ground */
			vel.z=0.9*fabs(vel.z);
			vel*=0.8; // ground drag
			pos.z=0.0;
			//printf("g");
		}
	}
};

/* An object that can push or pull on particles. */
class particlePusher {
public:
	virtual void simulate(double dt,std::vector<particle> &p,physics::library &lib) =0;
	virtual void draw(std::vector<particle> &p,physics::library &lib) =0;
};


/* A chain/rope segment */
class chainSegment : public particlePusher {
	int s,e; /* vertex numbers (in particle list) */
	float max_length;
public:
	
	chainSegment(int s_,int e_,std::vector<particle> &p)
		:s(s_), e(e_)
	{
		max_length=length(p[e].pos-p[s].pos);
	}
	
	
	void simulate(double dt,std::vector<particle> &p,physics::library &lib) {
		/* Total distance interaction */
		vec3 along=p[e].pos-p[s].pos;
		float len=length(along);
		float tension=0.0;
		if (len > max_length) {
			tension=(len-max_length);
			// p[e].pos=p[s].pos+max_length*normalize(along); // hard constaint
			
			float spring_constant = 1.0e6; /* newtons of force per meter of stretch */
			vec3 myforce=spring_constant*tension*along;
			p[e].netforce-=myforce;
			p[s].netforce+=myforce;
		}
	}
	
	void draw(std::vector<particle> &p,physics::library &lib) {
		vec3 spos=p[s].pos, epos=p[e].pos;
		
		glLineWidth(30.0);
		/* Draw segment, then shadow */
		for (int shadow=0;shadow<=1;shadow++) {
		  if (shadow) {
		  	glDisable(GL_TEXTURE_2D);
			glDepthMask(GL_FALSE);
			glColor4f(0.0f,0.0f,0.0f,0.3f);
			spos.z=epos.z=0.0; /* squish shadows down */
		  } else { /* ordinary chain */
			glColor3f(1,0,0);
		  }
		  glBegin(GL_LINES); //<- move outside draw call?
		  glVertex3fv(spos); glVertex3fv(epos);
		  glEnd();
		}
		glDepthMask(GL_TRUE);
	}
};



/* A sphere-shaped pusher */
class spherePusher : public particlePusher {
public:
	vec3 location;
	float radius;
	vec4 color;
	spherePusher() {
		location=vec3(0.0);
		radius=1.0;
		color=vec4(1,0,0,0.5);
	}
	
	virtual void push(const vec3 &direction,std::vector<particle> &p) { /* ignore push */ }
	
	virtual void draw(std::vector<particle> &p,physics::library &lib)
	{
		int slices=8,stacks=5;
		for (int pass=1;pass<=3;pass++) {
			glPushMatrix();
			if (pass==3) { glScalef(1,1,0); /* squish shadows onto ground */ }
			glTranslatef(VEC3_TO_XYZ(location));
			
			if (pass!=2) glDepthMask(GL_FALSE); /* transparent -> no depth writes */
			if (pass==1) { /* first pass: transparent outline */
				glColor4fv(color);
				glutSolidSphere(radius,slices,stacks);
			} 
			else if (pass==2) { /* second pass: wireframe */
				glColor4f(0,0,0,1.0);
				glLineWidth(1.0);
				glutWireSphere(radius,slices,stacks);
			} 
			else if (pass==3 && location.z>=0.0) { /* final pass: shadow */
				glColor4f(0,0,0,0.5);
				glutWireSphere(radius,slices,stacks);
				glColor4f(0,0,0,0.3);
				glutSolidSphere(radius,slices,stacks);
			}
			glPopMatrix();
			glDepthMask(GL_TRUE); /* back to normal */
		}
	}
};


/* A force-based particle-repelling sphere */
class softSphere : public spherePusher {
public:
	softSphere(vec3 location_,float radius_,vec4 color_=vec4(1,0,0,0.5)) {
		location=location_; radius=radius_; color=color_;
	}
	
	virtual void simulate(double dt,std::vector<particle> &p,physics::library &lib)
	{
		for (unsigned int i=0;i<p.size();i++)
		{
			vec3 fear=-(location-p[i].pos);
			if (length(fear)<1.0) { // close enough to scare me! 
				p[i].netforce+=normalize(fear)*1000.0;
			}
		}
	}
};

/* A constraint-based rolling ground sphere */
class rollySphere : public spherePusher {
public:
	int i; /* the particle I am connected to */
	rollySphere(int i_,vec4 color_,std::vector<particle> &p) {
		i=i_;
		location=p[i].pos;
		color=color_;
	}
	
	virtual void simulate(double dt,std::vector<particle> &p,physics::library &lib)
	{
		// Particle should follow our position
		p[i].netforce+=1.0e5*(location-p[i].pos);
		
		// Weakly, we should follow the particle's position
		location+=0.1*(p[i].pos-location);
		
		
		if (p[i].pos.z<radius) { /* below the ground */
			p[i].pos.z=radius;
			if (p[i].netforce.z<0.0) p[i].netforce.z=0.0; /* no downward forces can apply */
		}
	}
	
	virtual void push(const vec3 &direction,std::vector<particle> &p) 
	{ /* transmit push down to connected node */
		location+=direction; 
	}
};

/* A particle system */
class particlesystem : public physics::object {
	std::vector<particle> p;
	std::vector<particlePusher *> pushers;
	
	unsigned int cur_sphere; /* the sphere we're currently controlling with the camera */
	std::vector<spherePusher *> spheres;
	
public:
	particlesystem(void) 
		:physics::object(0.001)  /* <- our timestep, in seconds */
	{
		/* Make the chain */
		int nnode=100;
		for (int i=0;i<nnode;i++) {
			particle np;
			np.pos=vec3(10.0*(i-nnode/2)/nnode,0,2.0);
			np.vel=vec3(-2.0,-2.0,2.0); // (randVec3()-vec3(0.5));
			np.mass=1.0;
			p.push_back(np);
			
			if (i-1>=0)
				pushers.push_back(new chainSegment(i-1,i,p));
		}
		
		/* Make the spheres */
		spherePusher *s;
		
		s=new softSphere(vec3(0.0),1.0,vec4(0.0,1.0,0.0,0.4)); // green
		pushers.push_back(s); spheres.push_back(s);
		
		s=new rollySphere(0,vec4(1.0,0.0,0.0,0.2),p); // red
		pushers.push_back(s); spheres.push_back(s);
		
		s=new rollySphere(nnode/2,vec4(1.0,0.0,1.0,0.2),p); // purple
		pushers.push_back(s); spheres.push_back(s);
		
		cur_sphere=0;
	}
	
	void simulate(physics::library &lib) {
		float slowmo=1.0;
		float dt=timestep.dt*slowmo;
		
		/* Total up kinetic energy of particles */
		double total_energy=0.0;
		for (unsigned int i=0;i<p.size();i++) {
			total_energy+=0.5*p[i].mass*dot(p[i].vel,p[i].vel);
		}
		printf("System kinetic energy: %.1f J\n",total_energy);
		
		// Fix the position of the camera-controlled sphere
		vec3 camera_to_sphere=-camera_orient.z*3.0-camera_orient.y*0.4;
		if (lib.key['f']) { /* flip to next sphere */
			cur_sphere++;
			if (cur_sphere>=spheres.size()) cur_sphere=0;
			camera=spheres[cur_sphere]->location-camera_to_sphere;
			lib.key['f']=false;
		}
		spheres[cur_sphere]->location=camera+camera_to_sphere;
		
		// Start timesteps
		for (unsigned int i=0;i<p.size();i++)
		{
			p[i].start_step(dt);
			if (length(p[i].vel)>100000.0) {std::cout<<"Unstable at t="<<timestep.last_t<<"\n"; exit(1);}
		}
		
		// Loop over pushers, add in net forces to particles
		for (unsigned int i=0;i<pushers.size();i++) {
			pushers[i]->simulate(dt,p,lib);
		}
		
		// Loop over spheres, resolve collisions
		for (unsigned int i=0;i<spheres.size();i++) 
		for (unsigned int j=i+1;j<spheres.size();j++) {
			vec3 along=spheres[i]->location - spheres[j]->location;
			float len=length(along);
			float minlen=spheres[i]->radius+spheres[j]->radius;
			if (len<minlen) 
			{ /* these spheres are intersecting */
				spheres[i]->push( along*(minlen-len),p);
				spheres[j]->push(-along*(minlen-len),p);
			}
		}
		
		// Finish timestep
		for (unsigned int i=0;i<p.size();i++) {
			p[i].finish_step(dt);
		}
	}
	
	void draw(physics::library &lib) {
		// Draw all pushers
		for (unsigned int i=0;i<pushers.size();i++)
			pushers[i]->draw(p,lib);
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	if (!lib.key_toggle['f']) {
		//camera=vec3(0.0,-5.0,1.7);
		camera_orient.x=vec3(1,0,0);
		camera_orient.z=normalize(vec3(0,-0.6,0.5));
	}
	lib.world->add(new particlesystem());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* use SOIL ground texture */
	static GLuint texture=SOIL_load_OGL_texture(
		"ground.jpg",0,0,SOIL_FLAG_MIPMAPS|SOIL_FLAG_TEXTURE_REPEATS
	);
	glBindTexture(GL_TEXTURE_2D,texture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,32);
	/* Draw opaque white ground plane */
	int groundsize=500;
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glPushMatrix();glTranslatef(0,0,-0.01); /* <- shove down, to reduce Z fighting */
	glBegin(GL_QUADS);
	glTexCoord2i(-groundsize,-groundsize); glVertex2i(-groundsize,-groundsize);
	glTexCoord2i(+groundsize,-groundsize); glVertex2i(+groundsize,-groundsize);
	glTexCoord2i(+groundsize,+groundsize); glVertex2i(+groundsize,+groundsize);
	glTexCoord2i(-groundsize,+groundsize); glVertex2i(-groundsize,+groundsize);
	glEnd();
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);

	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Now draw objects */
	lib.world->draw(lib);
}


